##############################################################################
#
# Copyright (c) 2010 ViFiB SARL and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################


import os
import urllib
import posixpath
import base64
import hashlib
import json


def _get_hash_from_file(path, hash_object):
    """
      Calculate the hash from file.
    """
    f = open(path, 'rb')
    try:
        chunk = f.read(2**16)
        while chunk:
            hash_object.update(chunk)
            chunk = f.read(2**16)
        return hash_object.hexdigest()
    finally:
        f.close()


def _get_sha256_from_file(path):
    """
      Return the sha256sum from file.
    """
    if not os.path.isfile(path):
        raise ValueError, 'Could not extract sha256sum.' \
                                         ' The path is not a file: %s ' % path
    return _get_hash_from_file(path, hashlib.sha256())


def check_sha256sum(path, sha256sum):
    """Tell whether the SHA256sum checksum of the file at path matches
    """
    return sha256sum == _get_hash_from_file(path, hashlib.sha256())


def _update_network_cached_map(network_cache, sha256sum, url, logger):
    """
      If the url is not into the map file, it must be updated.
    """
    # Keep the log
    map_path = os.path.join(os.environ.get('PWD'), '.networkcached.log')
    mode = 'r+'
    if not os.path.exists(map_path):
        mode = 'w+'
    f = open(map_path, mode)
    try:
        flag = False
        for line in f:
            if url in line:
                flag = True
                break
        if not flag:
          f.seek(0, 2)
          f.write('%s %s\n' % (sha256sum, url))
          f.truncate(f.tell())
    finally:
        f.close()

    # Update the map at networkcached server.
    network_cached_url = os.path.join(network_cache, 'map/add')
    try:
        result = urllib.urlopen(network_cached_url, \
                     urllib.urlencode({'sha256sum': sha256sum, 'url': url}))
        if int(result.code) != 200:
            logger.info('Fail to update the network cache map: %s %s' % \
                                                         (sha256sum, url))
    except IOError, e:
         logger.info('An error occurred to update the map in networkcached.' \
                                                                ' %s' % str(e))


def get_sha256sum_from_networkcached(network_cache, url, logger):
    """
      Return the sha256sum if the url exists in networkcached server.
    """
    network_cached_url = os.path.join(network_cache,
                                          'map/get', base64.encodestring(url))
    try:
        result = urllib.urlopen(network_cached_url)
        if int(result.code) == 200:
          return json.loads(result.read().strip()).get('sha256sum', None)
        logger.info('The url is not cached yet: %s' % url)
    except IOError, e:
        logger.info('An error occurred to get sha256sum of url %s. %s' % \
                                                               (url, str(e)))


def download_network_cached(network_cache, path, url, logger):
    """Download from a network cache provider

    If something fail (providor be offline, or hash_string fail), we ignore
    network cached files.

    return True if download succeeded.
    """
    if network_cache in (None, '',):
        # Not able to use network cache
        return False

    sha256sum = get_sha256sum_from_networkcached(network_cache, url, logger)
    if sha256sum is None:
      return False

    network_cached_url = os.path.join(network_cache, sha256sum)
    logger.info('Downloading from network cache %s' % network_cached_url)
    try:
        path, headers = urllib.urlretrieve(network_cached_url, path)
        if not check_sha256sum(path, sha256sum):
            logger.info('MD5/SHA256 checksum mismatch downloading %r' % \
                                                           network_cached_url)
            return False
    except IOError, e:
        logger.info('Fail to download from network cache %s: %s' % \
                                                  (network_cached_url, str(e)))
        return False

    return True


def upload_network_cached(network_cache, external_url, path, logger):
    """Upload file to a network cache server"""
    if network_cache in [None, '']:
        logger.debug(
         'Upload cache ignored, network-cache was not provided')
        return False

    sha256sum = _get_sha256_from_file(path)
    try:
        f = open(path, 'r')
        data = f.read()
        url = os.path.join(network_cache, sha256sum)
        try:
            result = urllib.urlopen(url, urllib.urlencode({
                                         "data": base64.encodestring(data)}))
            if result.code == 200 and \
                        json.loads(result.read()).get('sha') == sha256sum:
                _update_network_cached_map(network_cache, sha256sum,
                                           external_url, logger)

        except (IOError, EOFError), e:
            logger.info('Fail to upload cache on %s. %s' % \
                                                      (external_url, str(e)))
    finally:
        f.close()
    return True


def get_filename_from_url(url):
    """Inspired how pip get filename from url.
    """
    url = url.split('#', 1)[0]
    url = url.split('?', 1)[0]
    url = url.rstrip('/')
    name = posixpath.basename(url)
    assert name, (
           'URL %r produced no filename' % url)
    return name
